/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

/* originally based on mod_speling.c
   Changes Copyright (c) 2003 Martin J. Du"rst, duerst@w3.org */

#include "apr.h"
#include "apr_file_io.h"
#include "apr_strings.h"
#include "apr_lib.h"
#include "apr_xlate.h"

#define APR_WANT_STRFUNC
#include "apr_want.h"

#define WANT_BASENAME_MATCH

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_request.h"
#include "http_log.h"

/* mod_fileiri.c - by Martin Du"rst <duerst@w3.org> June, 2003
 *
 * Started 2003-06-01, see CVS history at
 * http://dev.w3.org/cvsweb/apache-modules/mod_fileiri/mod_fileiri.c
 * for more information on how the code developed.
 */

/* mod_fileiri implements http IRIs for directories/files, i.e.
 * if accepts URIs with non-ASCII characters encoded in UTF-8 and
 * converts them to the legacy encoding used in the file system
 * (which can be specified per directory, or even finer if
 *  necessary (although that's a real hack)).
 *
 * What is more, it continues to accept requests in the legacy
 * encoding specified, and redirects them to the correct UTF-8
 * form, which then returns the actual document (without looping).
 *
 * There is also a backwards mode, which does redirects from
 * URIs in a specified legacy encoding to UTF-8 if the directory/
 * filenames are in UTF-8.
 */


module AP_MODULE_DECLARE_DATA fileiri_module;

typedef struct iriconfig {
    const char *setting;
    const char *file_charset;
    const char *oldfile_charset;
    const char *directory;
    struct iriconfig *prev; /* pointer to previous iriconfig */
} iriconfig;

/*
 * UTF-8 lead byte enum and table for tight checking
 * (all this utf-8 stuff probably should go into some library)
 */

typedef enum {
    end, /* end of string */
    ill, /* illegal byte */
    asc, /* us-ascii     */
    trl, /* trailing     */
    by2, /* two bytes    */
    e0,  /* three bytes, e0 lead byte */
    by3, /* three bytes, normal case  */
    ed,  /* three bytes, ed lead byte */ 
    p13, /* four  bytes, planes 1-3   */
    by4, /* four  bytes, normal case (planes 4-15) */
    p16, /* four  bytes, plane 16     */
} utf8lead;

static utf8lead UTF8lead[256] = {
    /* should all control codes be illegal? or allowed? */
    /* never make 0x00 legal, code depends on it being illegal */
    /* 0x00 */ end,  ill,  ill,  ill,  ill,  ill,  ill,  ill,
    /* 0x08 */ ill,  asc,  asc,  ill,  ill,  asc,  ill,  ill,
    /* 0x10 */ ill,  ill,  ill,  ill,  ill,  ill,  ill,  ill,
    /* 0x18 */ ill,  ill,  ill,  ill,  ill,  ill,  ill,  ill,
    /* 0x20 */ asc,  asc,  asc,  asc,  asc,  asc,  asc,  asc,
    /* 0x28 */ asc,  asc,  asc,  asc,  asc,  asc,  asc,  asc,
    /* 0x30 */ asc,  asc,  asc,  asc,  asc,  asc,  asc,  asc,
    /* 0x38 */ asc,  asc,  asc,  asc,  asc,  asc,  asc,  asc,
    /* 0x40 */ asc,  asc,  asc,  asc,  asc,  asc,  asc,  asc,
    /* 0x48 */ asc,  asc,  asc,  asc,  asc,  asc,  asc,  asc,
    /* 0x50 */ asc,  asc,  asc,  asc,  asc,  asc,  asc,  asc,
    /* 0x58 */ asc,  asc,  asc,  asc,  asc,  asc,  asc,  asc,
    /* 0x60 */ asc,  asc,  asc,  asc,  asc,  asc,  asc,  asc,
    /* 0x68 */ asc,  asc,  asc,  asc,  asc,  asc,  asc,  asc,
    /* 0x70 */ asc,  asc,  asc,  asc,  asc,  asc,  asc,  asc,
    /* 0x78 */ asc,  asc,  asc,  asc,  asc,  asc,  asc,  ill,
    /* 0x80 */ trl,  trl,  trl,  trl,  trl,  trl,  trl,  trl,
    /* 0x88 */ trl,  trl,  trl,  trl,  trl,  trl,  trl,  trl,
    /* 0x90 */ trl,  trl,  trl,  trl,  trl,  trl,  trl,  trl,
    /* 0x98 */ trl,  trl,  trl,  trl,  trl,  trl,  trl,  trl,
    /* 0xA0 */ trl,  trl,  trl,  trl,  trl,  trl,  trl,  trl,
    /* 0xA8 */ trl,  trl,  trl,  trl,  trl,  trl,  trl,  trl,
    /* 0xB0 */ trl,  trl,  trl,  trl,  trl,  trl,  trl,  trl,
    /* 0xB8 */ trl,  trl,  trl,  trl,  trl,  trl,  trl,  trl,
    /* 0xC0 */ ill,  ill,  by2,  by2,  by2,  by2,  by2,  by2,
    /* 0xC8 */ by2,  by2,  by2,  by2,  by2,  by2,  by2,  by2,
    /* 0xD0 */ by2,  by2,  by2,  by2,  by2,  by2,  by2,  by2,
    /* 0xD8 */ by2,  by2,  by2,  by2,  by2,  by2,  by2,  by2,
    /* 0xE0 */ e0,   by3,  by3,  by3,  by3,  by3,  by3,  by3,
    /* 0xE8 */ by3,  by3,  by3,  by3,  by3,  ed,   by3,  by3,
    /* 0xF0 */ p13,  by4,  by4,  by4,  p16,  ill,  ill,  ill,  
    /* 0x18 */ ill,  ill,  ill,  ill,  ill,  ill,  ill,  ill,
};

static int isUTF8 (unsigned char *p)
{
    /* figure out whether the input is UTF-8 */
    while (1) {
        switch (UTF8lead[*p++]) {
	  case end:
            return 1; /* clean end */
	  case ill:
	  case trl:
	    return 0;
	  case asc:
	    break;
	  case by2:
	    if (UTF8lead[*p++]!=trl)
		return 0;
	    break;
	  case e0 :
            if ((((*p++)&0xE0)!=0xA0) || (UTF8lead[*p++]!=trl))
	        return 0;
            break;
          case by3:
	    if ((UTF8lead[*p++]!=trl) || (UTF8lead[*p++]!=trl))
		return 0;
            break;
	  case ed :
            if ((((*p++)&0xE0)!=0x80) || (UTF8lead[*p++]!=trl))
	        return 0;
            break;
          case p13:
            if (((*p)<0x90 || 0xBF<(*p++)) 
	          || (UTF8lead[*p++]!=trl)
		  || (UTF8lead[*p++]!=trl))
	        return 0;
	    break;
	  case by4:
	    if ((UTF8lead[*p++]!=trl)
		  || (UTF8lead[*p++]!=trl)
                  || (UTF8lead[*p++]!=trl))
		return 0;
	    break;
	  case p16:
            if ((((*p++)&0xF0)!=0x80)
		  || (UTF8lead[*p++]!=trl)
		  || (UTF8lead[*p++]!=trl))
	        return 0;
	    break;
        }
    }
}

static int isASCII (unsigned char *p)
{
    while (*p)
        if (*p++ >= 0x80)
            return 0;
    return 1;
}


/* encoding conversion helper function */
/* for easy conversion of short pieces of text */
static char *transcode (apr_pool_t *pool, const char *from,
                        const char* to, char *in)
{
    apr_xlate_t *convset;
    apr_size_t inlength = strlen(in);
    /* factor 4: max expansion to/from UTF-8 */
    apr_size_t bufflength = inlength * 4;
    apr_size_t restlength = bufflength;
    char *out = apr_pcalloc(pool, bufflength+1);

    if (!from || !to)
        return NULL;
    if (APR_SUCCESS != apr_xlate_open (&convset, to, from, pool)) {
        /* should do some logging here */
        return NULL;    /* translation not available */
    }
    apr_xlate_conv_buffer (convset, in, &inlength, out, &restlength);
    if (inlength==0) {  /* conversion complete */
        out[bufflength-restlength] = 0;
        return out;
    }
    else                /* conversion incomplete */
        return NULL;
}


/*
 * Respond to callbacks to create/merge configuration records.
 * Create/merge a configuration specific to this module for
 * a server or directory location.
 */

static void *make_iri_conf(apr_pool_t *p)
{
    iriconfig *cfg = (iriconfig *)apr_pcalloc(p, sizeof(iriconfig));

    cfg->setting = NULL;  /* not set at all */
    cfg->file_charset = NULL;
    cfg->oldfile_charset = NULL;
    cfg->prev = NULL;
    return cfg;
}

static void *create_iri_serv_conf(apr_pool_t *p, server_rec *s)
{
    iriconfig *cfg = make_iri_conf(p);
    cfg->directory = "/";  /* directory always ends with slash */
    return cfg;
}

static void *create_iri_dir_conf(apr_pool_t *p, char *dir)
{
    iriconfig *cfg = make_iri_conf(p);
    cfg->directory = dir ? dir : "/";
    return cfg;
}

static void copy_iri_conf(iriconfig *base, iriconfig *over)
{   /* helper function */
    if (!over || !base)  return;
    if (!over->setting)  over->setting = base->setting;
    if (!over->file_charset)  over->file_charset = base->file_charset;
    if (!over->oldfile_charset)  over->oldfile_charset = base->oldfile_charset;
}

static void *merge_iri_conf(apr_pool_t *p, void *basev, void *overridesv)
{
    iriconfig *base = (iriconfig *) basev;
    iriconfig *over = (iriconfig *) overridesv;
    int bdl, odl;

    /* Trying to be as robust as possible, order we get called
       does not always seem to be as expected.
       Merging two lists; invariants:
       o longer directory paths first
       o no two directory paths of the same length
    */

    if (!base)  return  over;
    if (!over)  return  base;

    bdl = strlen(base->directory);
    odl = strlen(over->directory);

    if (bdl == odl) {
      /*check for directory equality and if not equal, do some logging*/
        copy_iri_conf(base, over);     
        over->prev = merge_iri_conf(p, (void*)base->prev, (void*)over->prev);
        copy_iri_conf(over->prev, over);
        return over;
    }
    else if (bdl < odl) {
        over->prev = merge_iri_conf(p, (void*)base, (void*)over->prev);
        copy_iri_conf(over->prev, over);
        return over;
    }
    else {  /* bdl > odl */
        base->prev = merge_iri_conf(p, (void*)base->prev, (void*)over);
        copy_iri_conf(base->prev, base);
	return base;
    }
}

/*
 * Handler for the FileIRI directive
 */
static const char *set_iri(cmd_parms *cmd, void *mconfig, const char *setting)
{
    ((iriconfig*)mconfig)->setting = setting;
    return NULL;
}

/*
 * Handler for the FilenameCharset directive
 */
static const char *set_file_charset(cmd_parms *cmd, void *mconfig,
                                    const char *charset)
{
    ((iriconfig*)mconfig)->file_charset = charset;
    return NULL;
}

/*
 * Handler for the OldFilenameCharset directive
 */
static const char *set_oldfile_charset(cmd_parms *cmd, void *mconfig,
                                    const char *charset)
{
    ((iriconfig*)mconfig)->oldfile_charset = charset;
    return NULL;
}

/*
 * Define the directives specific to this module.
 * This structure is referenced later by the 'module' structure.
 */
static const command_rec iri_cmds[] =
{
    AP_INIT_TAKE1("FileIRI",
                  set_iri,
                  NULL,
                  OR_OPTIONS,
                  "setting for fileiri module: On|Off|Backwards"),
    AP_INIT_TAKE1("FilenameCharset",
                  set_file_charset,
                  NULL,
                  OR_OPTIONS,
                  "character encoding (charset) of file names"),
    AP_INIT_TAKE1("OldFilenameCharset",
                  set_oldfile_charset,
                  NULL,
                  OR_OPTIONS,
                  "former character encoding (charset) of file names"),
    { NULL }
};


/* map a whole directory path (either way) */
/*
 * reverse mapper, accepting paths/files in legacy encoding
 * and (externally) redirecting them to UTF-8 names
 * (watch out carefully for internal redirects)
 */

/* define flag positions */
#define noUTF8    0  /* except with Backwards or nonconvertible */
#define noLegacy  1  /* except nonconvertible */

static int map_fileiri(request_rec *r)
{
    iriconfig *cfg;
    int urilen, blocklen, loops;
    char *flags, *urileft, *fileleft, *newActual, *newUTF8, *path;
    const char *ref, *block, *uriUTF8;
    request_rec *redirecter;

    /* This is a sub request - don't mess with it */
    if (r->main)
        return DECLINED;

    /* We only want to worry about GET/HEAD/... */
    /* mjd: we are more deterministic than mod_speling,
               so we might want to deal with PUT,... too */
    if (r->method_number != M_GET)
        return DECLINED;

    /* We've already got a file of some kind or another */
    if (r->proxyreq)  /* MJD: don't really know what proxyrec means, but...*/
        return DECLINED;

    if (isASCII(r->filename))  /* whole filename is ASCII, nothing to do */
        return DECLINED;

    /* Don't do anything if request has no slash or is "/" */
    if (ap_rind(r->filename,'/') == -1 || strcmp(r->uri, "/") == 0)
        return DECLINED;

    cfg = ap_get_module_config(r->per_dir_config, &fileiri_module);

    /* check internal redirect */
    redirecter = r;
    while ((redirecter=redirecter->prev))
        if (redirecter->handler
                && !strcmp(redirecter->handler, "fileiri-redirect"))
	    break;

    /* get data from redirecter */
    if (redirecter) {
        flags   = (char*)apr_table_get(redirecter->notes,"fileiri-flags");
        block   = apr_table_get(redirecter->notes,"fileiri-block");
        uriUTF8 = apr_table_get(redirecter->notes,"fileiri-UTF-8");
    }
    else {
        flags = "11";
        block = "";
        uriUTF8 = NULL;
    }
    flags = apr_pstrndup(r->pool, flags, 2); /* make copy so we can change */
    blocklen = strlen(block);
    /* check that nothing got changed between redirecter and here */
    if (strncmp(block, r->uri, blocklen))
        return DECLINED;

    /* file with noLegacy URI (i.e. already ext redirected), done */
    urilen = strlen(r->uri);
    if (r->finfo.filetype && flags[noLegacy]=='1' &&
            (urilen==blocklen||isUTF8(r->filename)))
        return DECLINED;

    /*
     * The request should get here looking like this:
     * r->uri: /uri-start/somebytes1/somebytes2/more
     * r->filename: /file-start/somebytes1/somebytes2 r->path_info: /more
     */

    /* Check to see if the 'more' part is the same, if not, quit */
    path = r->path_info ? r->path_info : "";
    urilen -= strlen(path);
    if (strcmp(path, r->uri + urilen))
        return DECLINED;

    /* initializations for loop */
    urileft = apr_pstrndup(r->pool, r->uri, urilen);
    fileleft= apr_pstrdup(r->pool, r->filename);
    newActual = newUTF8 = "";
    loops = 0;

    /* Now we loop back up the directory hierarchy */
    while (1) {
        int filoc, dirlen, piecelen;
        char *piece, *pieceUTF8, *pieceActual;

        /* if remainder is ASCII, stop converting (if loops!=0, not ASCII) */
        if (loops && isASCII(fileleft))
            break;

        /* check blocking (already dealt with before internal redirect) */
        if (urilen <= blocklen)
	    break;

	/* find the right config entry in the list */
        filoc = ap_rind(fileleft, '/');
	dirlen = strlen(cfg->directory) - 1;
	while (cfg && dirlen > filoc)
	    cfg = cfg->prev;       /* -1 for trailing slash of directory */
        if (!cfg)
            break; /* done converting, do whatever appropriate */
	if (strncmp(cfg->directory, fileleft, dirlen)) {
            /* log the directory mismatch */
            break;
	}

	piece = apr_pstrdup(r->pool, fileleft + filoc + 1);
        piecelen = strlen(piece);

        /* check that we still match */
        if (strcmp(piece, urileft + (urilen - piecelen))
                || '/' != urileft[urilen - piecelen - 1]) {
	    break;
        }

        /* cut things back */
        urileft [urilen - piecelen - 1] = 0;
        urilen -= piecelen + 1;
        fileleft[filoc] = 0;
        /* do not break out of while loop past this point! */

        /* try to convert */
        pieceUTF8 = pieceActual = NULL;
	if (isASCII(piece))  ; /* do nothing */
	else if (isUTF8(piece)) {
	    if (!strcmp(cfg->setting,"On"))
  	        pieceActual = transcode (r->pool, "UTF-8",
                                         cfg->file_charset, piece);
	}
        else { /* legacy */
  	    if (!strcmp(cfg->setting,"On"))
  	        pieceUTF8   = transcode (r->pool, cfg->file_charset,
					 "UTF-8", piece);
            if (!strcmp(cfg->setting,"Backwards"))
	        pieceActual = pieceUTF8 = transcode (r->pool,
                                   cfg->oldfile_charset, "UTF-8", piece);
	}

        /* complete copying and set flags */
	if (pieceUTF8) {         /* some legacy was converted to UTF-8 */
	    flags[noLegacy] = '0';
	    if (!pieceActual)
                pieceActual = piece;
        }
        else if (pieceActual) {  /* some UTF-8 was converted to legacy */
	    flags[noUTF8] = '0';
	    pieceUTF8 = piece;
	}
	else
	    pieceUTF8 = pieceActual = piece;

        /* put things together */
  	if (*newUTF8) {
	    newUTF8   = apr_pstrcat(r->pool, pieceUTF8, "/", newUTF8, NULL);
	    newActual = apr_pstrcat(r->pool, pieceActual, "/", newActual,NULL);
	}
        else {
 	    newActual = pieceActual;
            newUTF8   = pieceUTF8;
        }
	loops++;
    }

    /* sort out results */
    if (!loops || (flags[noUTF8]=='1' && flags[noLegacy]=='1'))
        return DECLINED; /* nothing converted, no action needed */
    else if (flags[noLegacy]=='1' || *path) {
        /* internal redirect, continue to follow directory
           hierarchy if all UTF-8 and On, or some path left */
        char *nuri;
        block   = apr_pstrcat(r->pool, urileft, "/", newActual, NULL);
        nuri    = apr_pstrcat(r->pool, block, r->path_info, NULL);
        uriUTF8 = apr_pstrcat(r->pool, uriUTF8 ? uriUTF8 : urileft,
                                              "/", newUTF8, NULL);

        ref = apr_table_get(r->headers_in, "Referer");
	ap_log_rerror(APLOG_MARK, APLOG_INFO, APR_SUCCESS,
		      r, 
		      ref ? "Encoding redirected: %s to %s from %s"
		          : "Encoding redirected: %s to %s",
		      r->uri, nuri, ref);

	r->filename = apr_pstrcat(r->pool, "redirect:", nuri, NULL);
	r->handler = "fileiri-redirect";
        apr_table_setn(r->notes, "fileiri-flags", flags);
        apr_table_setn(r->notes, "fileiri-block", block);
        apr_table_setn(r->notes, "fileiri-UTF-8", uriUTF8);
	return OK;
    }
    else {    /* external redirect only if we are at the end */
        char *nuri = ap_escape_uri(r->pool, apr_pstrcat(r->pool, urileft,
				         "/", newUTF8, r->path_info, NULL));
        ref = apr_table_get(r->headers_in, "Referer");
	if (r->parsed_uri.query)
	    nuri = apr_pstrcat(r->pool, nuri, "?", r->parsed_uri.query, NULL);

	/* external redirect, to let client know
	   that UTF-8 version is cannonical */
	apr_table_setn(r->headers_out, "Location",
		       ap_construct_url(r->pool, nuri, r));

	ap_log_rerror(APLOG_MARK, APLOG_INFO, APR_SUCCESS,
		      r, 
		      ref ? "Fixed encoding: %s to %s from %s"
		          : "Fixed encoding: %s to %s",
		      r->uri, nuri, ref);

	return HTTP_MOVED_PERMANENTLY;
    }
}


/* internal redirect handler, copied from mod_rewrite */
/* would be nice if there were out-of-the-box support
   for something like this */
static int fileiri_redirect(request_rec *r)
{
    if (strcmp(r->handler, "fileiri-redirect")) {
        return DECLINED;
    }
    if (strncmp(r->filename, "redirect:", 9) != 0) {
        return DECLINED;
    }
    ap_internal_redirect(apr_pstrcat(r->pool, r->filename+9,
				     r->args ? "?" : NULL, r->args, NULL), r);
    return OK;
}


static void iri_register_hooks(apr_pool_t *p)
{
    /* main mapper, from UTF-8 to legacy or back */
    ap_hook_fixups(map_fileiri,NULL,NULL,APR_HOOK_LAST);
    /* internal redirect handler (needs separate phase) */
    ap_hook_handler(fileiri_redirect,NULL,NULL,APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA fileiri_module =
{
    STANDARD20_MODULE_STUFF,
    create_iri_dir_conf,        /* create per-dir config */
    merge_iri_conf,             /* merge per-dir config */
    /* no need for separate methods for server conf */
    create_iri_serv_conf,       /* server config */
    merge_iri_conf,             /* merge server config */
    iri_cmds,                   /* command apr_table_t */
    iri_register_hooks		/* register hooks */
};
